"""Tests for greatday's FileRepo class."""

from __future__ import annotations

from pathlib import Path

from pytest import fixture

from greatday._repo import FileRepo
from greatday._todo import GreatTodo


TODO_LINES: list[str] = [
    "o foo | id:1",
    "o bar | id:2",
    "o baz | id:3",
]


@fixture(name="repo")
def file_repo_fixture(tmp_path: Path) -> FileRepo:
    """Returns a FileRepo populated with dummy data."""
    todo_file = tmp_path / "todos.txt"
    repo = FileRepo(todo_file)
    for line in TODO_LINES:
        todo = GreatTodo.from_line(line).unwrap()
        repo.add(todo).unwrap()
    return repo


def test_add_and_get(repo: FileRepo) -> None:
    """Tests the FileRepo's add() and get() methods."""
    assert len(repo.all().unwrap()) == len(TODO_LINES)

    ID = "100"
    todo = GreatTodo.from_line(f"o new todo | id:{ID}").unwrap()
    repo.add(todo).unwrap()
    assert len(repo.all().unwrap()) == len(TODO_LINES) + 1

    new_todo = repo.get(ID).unwrap()
    assert new_todo == todo


def test_remove(repo: FileRepo) -> None:
    """Tests the FileRepo.remove() method."""
    ID = "1"
    removed_todo = repo.remove(ID).unwrap()
    all_todos = repo.all().unwrap()
    assert len(all_todos) == len(TODO_LINES) - 1
    assert removed_todo not in all_todos



"""Contains all database utilities."""

from __future__ import annotations

from functools import lru_cache as cache
import os
from typing import Any

from sqlalchemy.future import Engine
from sqlmodel import SQLModel, create_engine as sqlmodel_create_engine


@cache
def create_cached_engine(url: str, /, **kwargs: Any) -> Engine:
    """Helper function for creating a new (if necessary) sqlalchemy engine."""
    engine = create_engine(url, **kwargs)
    return engine


def create_engine(url: str, /, **kwargs: Any) -> Engine:
    """Wrapper around sqlmodel.create_engine() that makes sure tables exist."""
    if "echo" not in kwargs and "ECHO_SQL_QUERIES" in os.environ:
        kwargs["echo"] = True

    engine = sqlmodel_create_engine(url, **kwargs)
    SQLModel.metadata.create_all(engine)
    return engine



"""Greatday date utilities."""

from __future__ import annotations

from dataclasses import dataclass
import datetime as dt
from typing import Final

from dateutil.relativedelta import relativedelta
import magodo


MONDAY: Final = 0
TUESDAY: Final = 1
WEDNESDAY: Final = 2
THURSDAY: Final = 3
FRIDAY: Final = 4
SATURDAY: Final = 5
SUNDAY: Final = 6

# metatags (i.e. key-value tags) that accept relative date strings (e.g. '1d')
RELATIVE_DATE_METATAGS: Final = ["snooze", "until", "due"]


@dataclass(frozen=True)
class DateRange:
    """Represents a range of dates."""

    start: dt.date
    end: dt.date | None = None

    @classmethod
    def from_strings(cls, start_str: str, end_str: str = None) -> DateRange:
        """Constructs a DateRange from two strings."""
        start = magodo.dates.to_date(start_str)
        end = magodo.dates.to_date(end_str) if end_str else None
        return cls(start, end)


def get_relative_date(
    spec: str, *, start_date: dt.date = None, past: bool = False
) -> dt.date:
    """Converts `spec` to a timedelta and adds it to `date`.

    Args:
        spec: A timedelta specification string (e.g. '1d', '2m', '3y',
          'weekdays').
        start_date: The return value is a function of this argument and the
          timedelta constructed from `spec`. Defaults to today's date.
        past: If set, we use a relative date from the past instead of the
          future (e.g. '1d' will yield yesterday's date instead of today's).

    Examples:
        # Imports
        >>> import datetime as dt

        # Helper Functions
        >>> to_date = lambda x: dt.datetime.strptime(x, "%Y-%m-%d")
        >>> from_date = lambda x: x.strftime("%Y-%m-%d")
        >>> grd = lambda x, y: from_date(
        ...   get_relative_date(x, start_date=to_date(y))
        ... )
        >>> past_grd = lambda x, y: from_date(
        ...   get_relative_date(x, start_date=to_date(y), past=True)
        ... )

        # Default start date.
        >>> D = "2000-01-31"

        # Tests
        >>> grd("7d", D)
        '2000-02-07'

        >>> grd("7D", D)
        '2000-02-07'

        >>> grd("1m", D)
        '2000-02-29'

        >>> grd("1m", "2001-01-31")
        '2001-02-28'

        >>> grd("2M", D)
        '2000-03-31'

        >>> grd("3m", D)
        '2000-04-30'

        >>> grd("20y", D)
        '2020-01-31'

        >>> grd("weekdays", "2022-02-11")
        '2022-02-14'

        >>> past_grd("1d", D)
        '2000-01-30'
    """
    spec = spec.lower()
    if start_date is None:
        start_date = dt.date.today()

    delta: dt.timedelta | relativedelta
    if spec == "weekdays":
        weekday = start_date.weekday()
        days = {FRIDAY: 3, SATURDAY: 2}.get(weekday, 1)
        delta = dt.timedelta(days=days)
    else:
        ch = spec[-1]
        N = int(spec[:-1])

        if ch == "d":
            delta = dt.timedelta(days=N)
        elif ch == "m":
            delta = relativedelta(months=N)
        else:
            assert ch == "y"
            delta = relativedelta(years=N)

    if past:
        return start_date - delta
    else:
        return start_date + delta


def dt_from_date_and_hhmm(date: dt.date, hhmm: str) -> dt.datetime:
    """Given a date and a string of the form HHMM, construct a datetime."""
    spec = f"{date.year}-{date.month}-{date.day} {hhmm}"
    result = dt.datetime.strptime(spec, "%Y-%m-%d %H%M")
    return result


def matches_date_fmt(spec: str) -> bool:
    """Returns True iff spec matches the magodo date format.."""
    return len(spec) == 10 and spec.count("-") == 2


def matches_relative_date_fmt(spec: str) -> bool:
    """Returns True iff spec appears to be a relative date (e.g. 1d)."""
    return (
        len(spec) > 1
        and spec[:-1].isdigit()
        and spec[-1].lower() in ["d", "m", "y"]
    )


def to_great_date(spec: str, past: bool = False) -> dt.date:
    """Converts a date string into a date.

    Args:
        spec: The date string specification (use a supported date format).
        past: Treat relative dates (e.g. when `spec == "1d"`) as dates in
          the past instead of the future.

    NOTE: `spec` must match a date string specification supported by
    greatday (e.g. 'YYYY-MM-DD').
    """
    if matches_date_fmt(spec):
        return magodo.dates.to_date(spec)
    else:
        assert matches_relative_date_fmt(spec)
        return get_relative_date(spec, past=past)


def get_date_range(spec: str) -> DateRange:
    """Constructs a date range from a `spec`.

    Args:
        spec: date specification which MUST use a format of START:END where
          START and END are valid date specs (e.g. `2000-01-01`; '1d'; '5m:0d').

    Examples:
        # setup
        >>> a = "2000-01-01"
        >>> b = "2000-01-31"

        # tests
        >>> a_range = get_date_range(a)
        >>> a_range.start
        datetime.date(2000, 1, 1)
        >>> a_range.end is None
        True

        >>> ab_range = get_date_range(f"{a}:{b}")
        >>> ab_range.start
        datetime.date(2000, 1, 1)
        >>> ab_range.end
        datetime.date(2000, 1, 31)
    """
    start_and_end = [to_great_date(x, past=True) for x in spec.split(":")]
    if len(start_and_end) > 1:
        assert len(start_and_end) == 2
        start, end = start_and_end
    else:
        start = start_and_end[0]
        end = None

    return DateRange(start, end)



"""Tests for the GreatSession potoroo.UnitOfWork class."""

from __future__ import annotations

from pathlib import Path
from typing import Callable, Final

import metaman
from pytest import mark

from greatday._repo import SQLRepo
from greatday._session import GreatSession
from greatday._tag import GreatTag

from . import common as c


params = mark.parametrize

FakeUserValidator = Callable[[SQLRepo], bool]
FakeEditorUser = Callable[[Path], FakeUserValidator]
FAKE_EDITOR_USERS: list[FakeEditorUser] = []
fake_editor_user = metaman.register_function_factory(FAKE_EDITOR_USERS)


@fake_editor_user
def delete_one_todo(path: Path) -> FakeUserValidator:
    """Deletes one todo from the file."""
    ID: Final = "1"

    lines = path.read_text().split("\n")
    new_text = "\n".join(line for line in lines if f" id:{ID}" not in line)
    path.write_text(new_text)

    def validator(repo: SQLRepo) -> bool:
        assert len(repo.all().unwrap()) == len(c.TODO_LINES) - 1
        assert repo.get(ID).unwrap() is None
        return True

    return validator


@fake_editor_user
def add_one_todo(path: Path) -> FakeUserValidator:
    """Adds one new todo to file."""
    ID: Final = str(int(c.TODO_LINE_IDS[-1]) + 1)
    DESC: Final = "NEW TODO"

    with path.open("a") as f:
        f.write(f"o {DESC}\n")

    def validator(repo: SQLRepo) -> bool:
        assert len(repo.all().unwrap()) == len(c.TODO_LINES) + 1
        todo = repo.get(ID).unwrap()
        assert todo is not None
        assert DESC in todo.to_line()
        return True

    return validator


@fake_editor_user
def edit_one_todo(path: Path) -> FakeUserValidator:
    """Edits one todo in file."""
    ID: Final = "1"
    CTX: Final = "ONE"

    lines = path.read_text().split("\n")
    new_lines: list[str] = []
    old_todo_line: str | None = None
    for line in lines:
        if f" id:{ID}" in line:
            old_todo_line = line
            new_lines.append(line + f" @{CTX}")
        else:
            new_lines.append(line)

    path.write_text("\n".join(new_lines))

    def validator(repo: SQLRepo) -> bool:
        todo = repo.get(ID).unwrap()
        assert todo is not None
        assert f" @{CTX} " in todo.desc
        assert CTX in todo.contexts
        assert old_todo_line is not None
        assert todo.to_line().replace(f" @{CTX}", "") == old_todo_line
        return True

    return validator


@params("faker", FAKE_EDITOR_USERS)
def test_fake_editor_users(sql_repo: SQLRepo, faker: FakeEditorUser) -> None:
    """Tests all fake editor user functions registered above."""
    tag = GreatTag.from_query("")
    with GreatSession(sql_repo.url, tag) as session:
        validator = faker(session.path)
        session.commit()
        assert validator(sql_repo)



"""This file contains shared fixtures and pytest hooks.

https://docs.pytest.org/en/6.2.x/fixture.html#conftest-py-sharing-fixtures-across-multiple-files
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Iterator

from freezegun import freeze_time
from pytest import fixture

from greatday import db
from greatday.__main__ import main as gtd_main
from greatday._repo import SQLRepo
from greatday._todo import GreatTodo

from . import common as c


if TYPE_CHECKING:  # fixes pytest warning
    from clack.pytest_plugin import MakeConfigFile


pytest_plugins = ["clack.pytest_plugin"]


@fixture
def main(make_config_file: MakeConfigFile, tmp_path: Path) -> c.MainType:
    """Returns a wrapper around greatday's main() function."""

    data_dir = tmp_path / "data"
    database_url = "sqlite://"

    def inner_main(*args: str, **kwargs: Any) -> int:
        if "data_dir" not in kwargs:
            kwargs["data_dir"] = data_dir

        if "database_url" not in kwargs:
            kwargs["database_url"] = database_url

        cfg_kwargs = {k: str(v) for (k, v) in kwargs.items()}

        config_file = make_config_file("greatday_test_config", **cfg_kwargs)
        argv = ["greatday", "-c", str(config_file.path)] + list(args)
        return gtd_main(argv)

    return inner_main


@fixture(autouse=True, scope="session")
def frozen_time() -> Iterator[None]:
    """Freeze time until our tests are done running."""
    with freeze_time(f"{c.TODAY}T{c.hh}:{c.mm}:00.123456Z"):
        yield


@fixture
def sql_repo() -> Iterator[SQLRepo]:
    """SQLRepo pytext fixture

    Yields:
        A SQLRepo populated with dummy data.
    """
    url: Final = "sqlite://"
    key: Final = "DATABASE_URL"

    db.create_cached_engine.cache_clear()

    default_url = os.environ.setdefault(key, url)
    repo = SQLRepo(url)
    for line in c.TODO_LINES:
        todo = GreatTodo.from_line(line).unwrap()
        repo.add(todo).unwrap()

    yield repo

    if url == default_url:
        del os.environ[key]



"""Test utils used by multiple modules."""

from __future__ import annotations

from typing import Any, Final, Protocol


# datetime info used by freezegun (and tests)
hh: Final = "00"
mm: Final = "00"
hhmm: Final = f"{hh}{mm}"

YYYY: Final = "2000"
MM: Final = "01"
DD: Final = "01"
TODAY: Final = f"{YYYY}-{MM}-{DD}"

TOMORROW: Final = f"{YYYY}-{MM}-02"

# dummy todo lines.. used as test data
TODO_LINES = (
    # ID #1
    "o 2000-01-01 Do some Laundry | @home @boring foo:bar",
    # ID #2
    "(B) 2000-02-03 Buy groceries | @out @boring +buy foo:bar due:2000-02-03",
    # ID #3
    "x 2010-01-02 2010-01-01 Finish greatday tests | @dev +greatday"
    " mile:2 p:0",
    # ID #4
    "o 1900-01-01 Finish greatday tests | @dev +greatday due:2000-01-01",
    # ID #5
    "x 2022-06-05 Some other todo | @misc p:1",
)

# the database IDs that should be associated with each of the todo lines above
TODO_LINE_IDS = tuple(str(n) for n in range(1, len(TODO_LINES) + 1))

# Parameters used by test_get_by_tag() test function.
#
# query -> ids
#
# query: used to construct GreatTag objects
# ids: iist of todo line IDs that this query should match
QUERY_TO_TODO_IDS: list[tuple[str, list[int]]] = [
    ("", [1, 2, 3, 4, 5]),
    ("o", [1, 2, 4]),
    ("x", [3, 5]),
    ("@home", [1]),
    ("!@home", [2, 3, 4, 5]),
    ("!@home @boring", [2]),
    ("@home @boring", [1]),
    ("@boring", [1, 2]),
    ("+greatday", [3, 4]),
    ("+buy @boring", [2]),
    ("(a)", []),
    ("(b)", [2]),
    ("(a-b)", [2]),
    ("(a,b)", [2]),
    ("due", [2, 4]),
    ("!due", [1, 3, 5]),
    ("due=2000-01-01", [4]),
    ("due>=2000-01-01", [2, 4]),
    ("due>2000-01-01", [2]),
    ("p", [3, 5]),
    ("p>0", [5]),
    ("^2000-01-01", [1]),
    ("^2000-01-01:2010-12-31", [1, 2, 3]),
    ("^2000-01-01:2010-12-31 $2010-01-01", []),
    ("^2000-01-01:2010-12-31 $2010-01-02", [3]),
    ('"some"', [1, 5]),
    ('!"some"', [2, 3, 4]),
    ('"Some"', [5]),
    ('!"Some"', [1, 2, 3, 4]),
    ('c"some"', [1]),
    ('!c"some"', [2, 3, 4, 5]),
    ("id=2", [2]),
    ("id>2", [3, 4, 5]),
    ("@home | @out", [1, 2]),
]


class MainType(Protocol):
    """Type returned by main() fixture."""

    def __call__(self, *args: str, **kwargs: Any) -> int:
        """The signature of the main() function."""



"""Tests for greatday magodo spells."""

from __future__ import annotations

from eris import Err
from pytest import mark

from greatday._todo import GreatTodo

from . import common as c


params = mark.parametrize


@params(
    "in_line, out_line",
    [
        # --- x_points spell
        (
            "x:1 foo",
            f"x:{c.hhmm} {c.TODAY} {c.TODAY} foo | ctime:{c.hhmm} p:1",
        ),
        (
            "x:1 2022-06-07 foo",
            f"x:{c.hhmm} {c.TODAY} 2022-06-07 foo | ctime:{c.hhmm} p:1",
        ),
        (
            "x:1 2022-06-07 2022-06-06 foo",
            f"x:{c.hhmm} 2022-06-07 2022-06-06 foo | ctime:{c.hhmm} p:1",
        ),
        # --- snooze spell
        (
            "o foo snooze:1d",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TOMORROW}",
        ),
        (
            f"(P) foo due:{c.TODAY} snooze:1d",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TOMORROW}",
        ),
        (
            f"o foo due:{c.TODAY} s:1d",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TOMORROW}",
        ),
        # --- relative dates spell
        (
            "o foo due:0d",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TODAY}",
        ),
        (
            "o foo due:1d",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TOMORROW}",
        ),
        # --- due context spell
        (
            "o foo @due",
            f"o {c.TODAY} foo | ctime:{c.hhmm} due:{c.TODAY}",
        ),
        # --- appt spell
        (
            "o foo appt:0100",
            f"o {c.TODAY} foo | appt:0100 ctime:{c.hhmm}",
        ),
        (
            f"o foo due:{c.TODAY} appt:0100",
            f"(T) {c.TODAY} foo | appt:0100 ctime:{c.hhmm} due:{c.TODAY}",
        ),
        (
            f"o foo due:{c.TODAY} appt:0030",
            f"(C) {c.TODAY} foo | appt:0030 ctime:{c.hhmm} due:{c.TODAY}",
        ),
        # --- in-progress priority spell
        (
            "(D) foo",
            f"(D) {c.TODAY} foo | ctime:{c.hhmm} start:{c.hhmm}",
        ),
        (
            "o foo start:1234",
            f"o {c.TODAY} foo | ctime:{c.hhmm}",
        ),
    ],
)
def test_spells(in_line: str, out_line: str) -> None:
    """Tests intended to flex greatday's custom magodo spells."""
    todo_result = GreatTodo.from_line(in_line)
    assert not isinstance(todo_result, Err)

    todo = todo_result.ok()
    assert todo.to_line() == out_line



"""Custom types used by greatday."""

from typing import Any, Literal, Protocol, TypeVar

from magodo.types import AbstractTodo
from sqlalchemy.future import Engine


T = TypeVar("T", bound=AbstractTodo)

YesNoDefault = Literal["n", "default", "y"]


class CreateEngineType(Protocol):
    """The type of a `db.create_engine()` callable."""

    def __call__(self, url: str, /, **kwargs: Any) -> Engine:
        """The function's call signature."""



"""Common code used throughout this package."""

from __future__ import annotations

from functools import lru_cache as cache, partial
from typing import Callable, Final, List, cast

from magodo.types import Priority
from typist import literal_to_list


# special contexts
CTX_FIRST: Final = "FIRST"
CTX_INBOX: Final = "INBOX"
CTX_LAST: Final = "LAST"

# special ID type (assigned when no real ID exists)
NULL_ID: Final = "null"

# all valid todo lines must start with one of...
TODO_PREFIXES: Final = ("x ", "x:", "o ")


def drop_words(
    desc: str,
    *bad_words: str,
    op: Callable[[str, str], bool] = lambda x, y: x == y,
) -> str:
    """Removes all `bad_words` from the todo description `desc`."""
    desc_words = desc.split(" ")
    new_desc_words = []
    for word in desc_words:
        if not any(op(word, bad_word) for bad_word in bad_words):
            new_desc_words.append(word)
    return " ".join(new_desc_words)


def _startswith_op(x: str, y: str) -> bool:
    """Used as the value for the 'op' kwarg of `drop_word_from_desc()`."""
    return x.startswith(y)


drop_word_if_startswith = partial(drop_words, op=_startswith_op)


@cache
def todo_prefixes() -> tuple[str, ...]:
    """Returns all valid todo prefixes."""
    result = list(TODO_PREFIXES)
    for P in cast(List[str], literal_to_list(Priority)):
        result.append(f"({P}) ")
    return tuple(result)



"""Contains greatday's SQL model class definitions."""

# WARNING: Don't bother importing __future__.annotations in this module!
import datetime as dt
from typing import List, Optional

from sqlmodel import Column, Field, Relationship, SQLModel, String
from sqlmodel.sql.expression import Select, SelectOfScalar


# HACK: see https://github.com/tiangolo/sqlmodel/issues/189
Select.inherit_cache = True  # type: ignore[attr-defined]
SelectOfScalar.inherit_cache = True  # type: ignore[attr-defined]


###############################################################################
# abstract model classes
###############################################################################
class Base(SQLModel):
    """Abstract base model class."""

    id: Optional[int] = Field(default=None, primary_key=True)


class TodoLink(SQLModel):
    """Abstract model for association/link models."""

    todo_id: Optional[int] = Field(
        default=None, foreign_key="todo.id", primary_key=True
    )


class Tag(Base):
    """Abstract model class for todo.txt tags."""

    name: str = Field(sa_column=Column(String, unique=True))


###############################################################################
# link models (i.e. assocation tables for many-to-many relationships)
###############################################################################
class ProjectLink(TodoLink, table=True):
    """Association model for todos-to-projects relationships."""

    project_id: Optional[int] = Field(
        default=None, foreign_key="project.id", primary_key=True
    )


class ContextLink(TodoLink, table=True):
    """Association model for todos-to-contexts relationships."""

    context_id: Optional[int] = Field(
        default=None, foreign_key="context.id", primary_key=True
    )


class EpicLink(TodoLink, table=True):
    """Association model for todos-to-epics relationships."""

    epic_id: Optional[int] = Field(
        default=None, foreign_key="epic.id", primary_key=True
    )


class MetatagLink(TodoLink, table=True):
    """Association model for todos-to-metatags relationships."""

    metatag_id: Optional[int] = Field(
        default=None, foreign_key="metatag.id", primary_key=True
    )

    todo: "Todo" = Relationship(back_populates="metatag_links")
    metatag: "Metatag" = Relationship(back_populates="links")

    value: str


###############################################################################
# model used to store todos
###############################################################################
class Todo(Base, table=True):
    """Model class for greatday Todos."""

    # table columns
    create_date: dt.date
    desc: str
    done: bool
    done_date: Optional[dt.date]
    priority: str

    # relationships
    contexts: List["Context"] = Relationship(
        back_populates="todos", link_model=ContextLink
    )
    epics: List["Epic"] = Relationship(
        back_populates="todos", link_model=EpicLink
    )
    projects: List["Project"] = Relationship(
        back_populates="todos", link_model=ProjectLink
    )
    metatag_links: List["MetatagLink"] = Relationship(back_populates="todo")


###############################################################################
# tag models
###############################################################################
class Project(Tag, table=True):
    """Model class for todo.txt project tags (e.g. +greatday)."""

    todos: List[Todo] = Relationship(
        back_populates="projects", link_model=ProjectLink
    )


class Context(Tag, table=True):
    """Model class for todo.txt context tags (e.g. @home)."""

    todos: List[Todo] = Relationship(
        back_populates="contexts", link_model=ContextLink
    )


class Epic(Tag, table=True):
    """Model class for (magodo extended) todo.txt epic tags (e.g. #gtd)."""

    todos: List[Todo] = Relationship(
        back_populates="epics", link_model=EpicLink
    )


class Metatag(Tag, table=True):
    """Model class for metadata tags (e.g. due:2022-06-01)."""

    links: List[MetatagLink] = Relationship(back_populates="metatag")



"""Functions / classes used to create greatday's TUI."""

from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import Any, Final, Sequence

from potoroo import TaggedRepo
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text
from textual.app import App
from textual.widgets import Footer, Header, Static
from textual_inputs import TextInput
from vimala import vim

from ._common import CTX_FIRST, CTX_INBOX, CTX_LAST
from ._repo import SQLRepo
from ._session import GreatSession
from ._tag import GreatTag
from ._todo import GreatTodo


# HACK: Used to fix action parameter parenthesis bug (see PR:textual#562).
FAKE_RIGHT_PAREN: Final = "]]]"

# characters that should be removed from query names in most cases
BAD_QUERY_NAME_CHARS: Final = "() 0123456789\n"

# important/saved GreatLang queries
INBOX_QUERY: Final = f"o @{CTX_INBOX}"
FIRST_QUERY: Final = f"o due<=0d @{CTX_FIRST} | $0d @{CTX_FIRST}"
LAST_QUERY: Final = f"o due<=0d @{CTX_LAST} | $0d @{CTX_LAST}"
LATE_QUERY: Final = f"o due<0d @{CTX_LAST} | $0d @{CTX_LAST} due<0d"
TODAY_QUERY: Final = f"o due<=0d !@{CTX_FIRST} !@{CTX_LAST} | $0d p>0 | (d)"

# a mapping of name->query that will be displayed in the "Stats" textual panel
STATS_QUERY_MAP: dict[str, str] = {
    "(0) inbox": INBOX_QUERY,
    "(1) first": FIRST_QUERY,
    "(2) late": LATE_QUERY,
    "(3) today": TODAY_QUERY,
    "(4) last": LAST_QUERY,
}


class GreatHeader(Header):
    """Override the default Header for Styling"""

    def __init__(self) -> None:
        super().__init__()
        self.tall = False
        self.style = Style(color="white", bgcolor="rgb(98,98,98)")

    def render(self) -> Table:
        """Returns renderable header."""
        header_table = Table.grid(padding=(0, 1), expand=True)
        header_table.add_column(justify="left", ratio=0, width=8)
        header_table.add_column("title", justify="center", ratio=1)
        header_table.add_column("clock", justify="right", width=8)
        header_table.add_row(
            "🔤", self.full_title, self.get_clock() if self.clock else ""
        )
        return header_table


class GreatFooter(Footer):
    """Override the default Footer for Styling"""

    def make_key_text(self) -> Text:
        """Create text containing all the keys."""
        text = Text(
            style="white on rgb(98,98,98)",
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )
        for binding in self.app.bindings.shown_keys:
            key = binding.key.upper() if len(binding.key) > 1 else binding.key
            key_display = (
                key if binding.key_display is None else binding.key_display
            )
            hovered = self.highlight_key == binding.key
            description = binding.description.strip(BAD_QUERY_NAME_CHARS)
            key_text = Text.assemble(
                (
                    f" {key_display} ",
                    "reverse" if hovered else "default on default",
                ),
                f" {description} ",
                meta={
                    "@click": f"app.press('{binding.key}')",
                    "key": binding.key,
                },
            )
            text.append_text(key_text)
        return text


class StatsWidget(Static):
    """Widget that shows Todo statistics."""

    def __init__(
        self,
        repo: TaggedRepo[str, GreatTodo, GreatTag],
        ctx: Context,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__("", *args, **kwargs)
        self.repo = repo
        self.ctx = ctx

        # If set, all saved query stats will reload on refresh. Otherwise, only
        # stats on the current query will refresh.
        self.do_full_refresh = False

        # saved query name -> query stats Text object
        #
        # A cache used to implement partial refreshes (i.e. when
        # `self.do_full_refresh` is False).
        self._text_cache: dict[str, Text] = {}

    def render(self) -> Panel:
        """Render the statistics widget."""
        stats_query_map = STATS_QUERY_MAP.copy()

        text = Text()
        max_name_size = max(
            len(name.strip()) for name in stats_query_map.keys()
        )
        if not any(
            query == self.ctx.query for query in stats_query_map.values()
        ):
            stats_query_map.update({"\n::": self.ctx.query})

        for name, query in stats_query_map.items():
            saved_q_matches_current_q = bool(self.ctx.query == query)
            if saved_q_matches_current_q:
                style = "bold italic blue"
            else:
                style = ""

            extra_text = self._text_cache.get(name)
            if (
                saved_q_matches_current_q
                or extra_text is None
                or self.do_full_refresh
            ):
                pretty_name = name.upper()
                spaces = ""
                if (size := len(pretty_name.strip())) < max_name_size + 1:
                    spaces += (max_name_size - size) * " "
                pretty_name += ":"
                pretty_name += spaces

                tag = GreatTag.from_query(query)
                todos = self.repo.get_by_tag(tag).unwrap()
                group = StatsGroup.from_todos(todos)

                extra_text = Text(
                    f"{pretty_name}   "
                    f"X({group.done_stats.count}.{group.done_stats.points}) + "
                    f"O({group.open_stats.count}.{group.open_stats.points}) = "
                    f"XO({group.all_stats.count}.{group.all_stats.points})\n"
                )
                self._text_cache[name] = extra_text
            extra_text.style = style
            text.append_text(extra_text)

        self.do_full_refresh = False
        return Panel(text, title="Statistics")


@dataclass
class StatsGroup:
    """Todo query stats group.

    Attributes:
        all_stats: Stats on all todos matching this query.
        done_stats: Stats on all completed todos matching this query.
        open_stats: Stats on all un-completed (i.e. not done) todos matching
          this query.
    """

    all_stats: Stats
    done_stats: Stats
    open_stats: Stats

    @classmethod
    def from_todos(cls, todos: Sequence[GreatTodo] | None) -> StatsGroup:
        """Constructs a StatsGroup from an iterable of todos (or None)."""
        if todos is not None:
            all_todos = todos
        else:
            all_todos = []

        all_count = len(all_todos)
        all_points = sum(
            int(todo.metadata.get("p" if todo.done else "xp", "0"))
            for todo in all_todos
        )

        done_todos = [todo for todo in all_todos if todo.done]
        done_count = len(done_todos)
        done_points = sum(
            int(todo.metadata.get("p", "0")) for todo in done_todos
        )

        open_todos = [todo for todo in all_todos if not todo.done]
        open_count = len(open_todos)
        open_points = sum(
            int(todo.metadata.get("xp", "0")) for todo in open_todos
        )

        all_stats = Stats(count=all_count, points=all_points)
        done_stats = Stats(count=done_count, points=done_points)
        open_stats = Stats(count=open_count, points=open_points)
        return cls(
            all_stats=all_stats, done_stats=done_stats, open_stats=open_stats
        )


@dataclass
class Stats:
    """Todo query stats.

    Attributes:
        count: number of todos matched by this query.
        points: Sum of points (i.e. 'p') or expected points (i.e. 'xp') for
          todos matching this query.
    """

    count: int
    points: int


@dataclass
class Context:
    """Mutable TUI Context.

    Used to preserve state after closing the GreatApp instance, running vim,
    and then opening a new GreatApp instance.

    Attributes:
        query: The active todo query string.
        edit_todos: After closing the TUI, should we open up vim to edit
          matching todos?
    """

    query: str
    edit_todos: bool = False


class GreatApp(App):
    """Textual TUI Application Class."""

    def __init__(
        self,
        *,
        repo: TaggedRepo[str, GreatTodo, GreatTag],
        ctx: Context,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self.repo = repo
        self.ctx = ctx

        self.input_widget = TextInput(name="input", value=self.ctx.query)
        self.input_widget.cursor = (
            "|",
            Style(
                color="black",
                blink=True,
                bold=True,
            ),
        )

        self.main_widget = Static(
            Panel(
                _todo_lines_from_query(self.repo, self.ctx.query),
                title="Main",
            ),
            name="main",
        )
        self.stats_widget = StatsWidget(self.repo, self.ctx)

    async def on_load(self) -> None:
        """Configure key bindings."""
        n = 0
        for name, query in STATS_QUERY_MAP.items():
            query = query.replace(")", FAKE_RIGHT_PAREN)
            description = f"{name.lstrip(BAD_QUERY_NAME_CHARS).upper()} Query"
            await self.bind(
                str(n), f"new_query('{query}')", description, show=False
            )
            n += 1

        await self.bind("escape", "change_mode('normal')", "Normal Mode")
        await self.bind("enter", "submit", "Submit")
        await self.bind("e", "edit", "Edit Todos")
        await self.bind("i", "change_mode('insert')", "Insert Mode")
        await self.bind("I", "clear_and_insert", "Clear and Insert")
        await self.bind("r", "refresh", "Refresh")
        await self.bind("q", "quit", "Quit")

    async def on_mount(self) -> None:
        """Configure layout."""
        # configure header and footer...
        await self.view.dock(GreatHeader(), edge="top")
        await self.view.dock(GreatFooter(), edge="bottom")

        # configure other widgets...
        await self.view.dock(self.input_widget, edge="bottom", size=5)
        await self.view.dock(self.stats_widget, edge="left", size=50)
        await self.view.dock(self.main_widget, edge="top")

    async def action_change_mode(self, mode: str) -> None:
        """Action to toggle to/from insert mode and other modes."""
        if mode == "insert":
            self.input_widget.title = "INSERT"
            self.input_widget.refresh()
            await self.input_widget.focus()
        elif mode == "normal":
            self.input_widget.title = ""
            self.input_widget.refresh()
            await self.main_widget.focus()
        else:
            raise AssertionError(f"Bad mode: {mode!r}")

    async def action_clear_and_insert(self) -> None:
        """Clears input bar and enters Insert mode."""
        self.input_widget.value = ""
        self.input_widget._cursor_position = 0
        await self.action_change_mode("insert")

    async def action_edit(self) -> None:
        """Edits todos which match the current todo query."""
        self.ctx.edit_todos = True
        await self.action_quit()

    async def action_new_query(self, query: str) -> None:
        """Execute a new todo query."""
        query = query.replace(FAKE_RIGHT_PAREN, ")")
        self.input_widget.value = query
        self.input_widget._cursor_position = len(query)
        await self.action_submit()

    async def action_refresh(self) -> None:
        """Full refresh of TUI (e.g. stats + main panel will reload)."""
        self.stats_widget.do_full_refresh = True
        await self.action_submit()

    async def action_submit(self) -> None:
        """Executes the current todo query shown in the input bar."""
        self.ctx.query = self.input_widget.value
        self.stats_widget.refresh()
        text = _todo_lines_from_query(self.repo, self.ctx.query)
        await self.main_widget.update(Panel(text, title="Todo List"))
        await self.action_change_mode("normal")


def _todo_lines_from_query(
    repo: TaggedRepo[str, GreatTodo, GreatTag], query: str
) -> str:
    tag = GreatTag.from_query(query)
    todos = repo.get_by_tag(tag).unwrap()

    result = ""
    for todo in sorted(todos):
        result += todo.to_line() + "\n"

    return result


def start_textual_app(db_url: str, *, verbose: int = 0) -> None:
    """Starts the TUI using the GreatApp class."""
    repo = SQLRepo(db_url)
    ctx = Context(TODAY_QUERY)
    run_app = partial(
        GreatApp.run,
        repo=repo,
        ctx=ctx,
        title="Greatday TUI",
        log="greatday_textual.log",
    )
    run_app()

    while ctx.edit_todos:
        tag = GreatTag.from_query(ctx.query)
        with GreatSession(db_url, tag, verbose=verbose) as session:
            vim(session.path).unwrap()
            session.commit()

        ctx.edit_todos = False
        run_app()



"""Test for greatday's 'list' subcommand."""

from __future__ import annotations

from _pytest.capture import CaptureFixture
from pytest import mark

from greatday._repo import SQLRepo

from . import common


params = mark.parametrize


@params("query,expected_ids", common.QUERY_TO_TODO_IDS)
def test_list(
    capsys: CaptureFixture,
    main: common.MainType,
    sql_repo: SQLRepo,
    query: str,
    expected_ids: list[int],
) -> None:
    """Tests the 'list' subcommand."""
    del sql_repo

    assert main("list", query) == 0
    captured = capsys.readouterr()
    id_list: list[int] = []
    for line in captured.out.split("\n"):
        line = line.strip()
        for word in line.split(" "):
            if word.startswith("id:"):
                _, ID = word.split(":")
                id_list.append(int(ID))

    id_list.sort()
    expected_ids.sort()
    assert id_list == expected_ids



"""Contains the Tag class."""

from __future__ import annotations

from dataclasses import dataclass, field
import enum
import string
from typing import Callable, Iterable, cast

from eris import ErisResult, Err, Ok
from logrus import Logger
import magodo
from magodo.types import Priority

from ._dates import (
    RELATIVE_DATE_METATAGS,
    DateRange,
    get_date_range,
    get_relative_date,
    matches_date_fmt,
    matches_relative_date_fmt,
)


logger = Logger(__name__)

QueryParser = Callable[[str], ErisResult[str]]


class MetatagOperator(enum.Enum):
    """Used to determine what kind of metatag constraint has been specified."""

    # exists / not exists
    EXISTS = enum.auto()
    NOT_EXISTS = enum.auto()

    # comparison operators
    EQ = enum.auto()
    NE = enum.auto()
    LT = enum.auto()
    LE = enum.auto()
    GT = enum.auto()
    GE = enum.auto()


class MetatagValueType(enum.Enum):
    """Specifies the data type of a MetatagFilter's value."""

    DATE = enum.auto()
    INTEGER = enum.auto()
    STRING = enum.auto()


@dataclass(frozen=True)
class MetatagFilter:
    """Represents a single metatag filter (e.g. 'due<=0d' or '!recur')."""

    key: str
    value: str = ""
    op: MetatagOperator = MetatagOperator.EXISTS
    value_type: MetatagValueType = MetatagValueType.STRING


class DescOperator(enum.Enum):
    """Used to determine the type of description constraint specified."""

    CONTAINS = enum.auto()
    NOT_CONTAINS = enum.auto()


@dataclass(frozen=True)
class DescFilter:
    """Represents a description query filter (e.g. '"foo"' or '!"bar"')."""

    value: str
    case_sensitive: bool | None = None
    op: DescOperator = DescOperator.CONTAINS


@dataclass(frozen=True)
class GreatTag:
    """A collection of `Tag`s that have been ORed together."""

    tags: Iterable[Tag]

    @classmethod
    def from_query(cls, query: str) -> GreatTag:
        """Build a GreatTag using a query string."""
        tags: list[Tag] = []
        for subquery in query.split(" | "):
            tag = Tag.from_query(subquery)
            tags.append(tag)

        return cls(tuple(tags))


@dataclass
class Tag:
    """Tag used to filter Todos."""

    contexts: list[str] = field(default_factory=list)
    create_date_ranges: list[DateRange] = field(default_factory=list)
    desc_filters: list[DescFilter] = field(default_factory=list)
    done_date_ranges: list[DateRange] = field(default_factory=list)
    done: bool | None = None
    epics: list[str] = field(default_factory=list)
    metatag_filters: list[MetatagFilter] = field(default_factory=list)
    priorities: list[Priority] = field(default_factory=list)
    projects: list[str] = field(default_factory=list)

    @classmethod
    def from_query(cls, query: str) -> Tag:
        """Build a Tag using a query string."""
        tag = cls()

        q = query
        while q:
            for parser in [
                tag.prefix_tag_parser_factory("#", "epics"),
                tag.prefix_tag_parser_factory("@", "contexts"),
                tag.prefix_tag_parser_factory("+", "projects"),
                tag.done_parser,
                tag.date_range_parser_factory("^", "create_date_ranges"),
                tag.date_range_parser_factory("$", "done_date_ranges"),
                tag.metatag_parser,
                tag.desc_parser_factory("'"),
                tag.desc_parser_factory('"'),
                tag.priority_parser,
            ]:
                q_result = parser(q)
                if isinstance(q_result, Err):
                    err = q_result.err()
                    logger.debug(
                        "Parser failed to find match.",
                        parser=parser.__name__,
                        error=err.to_json(),
                    )
                else:
                    q = q_result.ok()
                    break
            else:
                raise RuntimeError(
                    "No parsers are able to parse this query. |"
                    f" query={query!r}"
                )

        return tag

    def prefix_tag_parser_factory(self, ch: str, attr: str) -> QueryParser:
        """Factory for parsers that handle normal tags (e.g. project tags)."""

        def parser(query: str) -> ErisResult[str]:
            prop_list = getattr(self, attr)
            word, *rest = query.split(" ")

            if word.startswith(ch):
                logger.debug("Filter on property.", word=word)
                prop_list.append(word[1:])
            elif word.startswith(f"!{ch}"):
                logger.debug("Filter on negative property.", word=word)
                prop_list.append(f"-{word[2:]}")
            else:
                return Err(
                    "First word of query does not match required tag prefix."
                    f" | prefix={ch} word={word}",
                )

            return Ok(" ".join(rest))

        return parser

    def done_parser(self, query: str) -> ErisResult[str]:
        """Parser for 'done' status (e.g. 'o' for open, 'x' for done)."""
        word, *rest = query.split(" ")
        if word.lower() == "o":
            self.done = False
        elif word.lower() == "x":
            self.done = True
        else:
            return Err("Next token is not 'o' or 'x'.")

        return Ok(" ".join(rest))

    def date_range_parser_factory(self, ch: str, attr: str) -> QueryParser:
        """Factory for create/done date range tokens."""

        def parser(query: str) -> ErisResult[str]:
            word, *rest = query.split(" ")
            if not word.startswith(ch):
                return Err("Next token is not a date range.")

            date_ranges = getattr(self, attr)
            date_range = get_date_range(word[1:])
            date_ranges.append(date_range)

            logger.debug(
                "Filtering on date range.",
                prefix=ch,
                date_range=date_range,
            )
            return Ok(" ".join(rest))

        return parser

    def metatag_parser(self, query: str) -> ErisResult[str]:
        """Parser for metadata checks."""
        word, *rest = query.split(" ")
        if word.isalpha():
            self.metatag_filters.append(
                MetatagFilter(word, op=MetatagOperator.EXISTS)
            )
        elif word.startswith("!") and word[1:].isalpha():
            self.metatag_filters.append(
                MetatagFilter(word[1:], op=MetatagOperator.NOT_EXISTS)
            )
        else:
            for op_string, metatag_op in [
                ("<=", MetatagOperator.LE),
                (">=", MetatagOperator.GE),
                ("<", MetatagOperator.LT),
                (">", MetatagOperator.GT),
                ("!=", MetatagOperator.NE),
                ("=", MetatagOperator.EQ),
            ]:
                key_and_value_string = word.split(op_string)
                if len(key_and_value_string) != 2:
                    continue

                key, value_string = key_and_value_string

                value = value_string
                value_type = MetatagValueType.STRING
                if (
                    op_string in ["=", "!="]
                    and key not in RELATIVE_DATE_METATAGS
                ):
                    pass
                elif matches_date_fmt(value_string):
                    value_type = MetatagValueType.DATE
                elif matches_relative_date_fmt(value_string):
                    value = magodo.dates.from_date(
                        get_relative_date(value_string)
                    )
                    value_type = MetatagValueType.DATE
                elif value_string.isdigit():
                    value_type = MetatagValueType.INTEGER

                self.metatag_filters.append(
                    MetatagFilter(
                        key,
                        value=value,
                        op=metatag_op,
                        value_type=value_type,
                    )
                )
                break
            else:
                return Err("Next token is not a metadata check.")

        return Ok(" ".join(rest))

    def desc_parser_factory(self, quote: str) -> QueryParser:
        """Factory for parser that handles description tokens."""

        def parser(query: str) -> ErisResult[str]:
            desc_op = DescOperator.CONTAINS
            q = query
            if q.startswith(f"!{quote}") or q.startswith(f"!c{quote}"):
                q = q[1:]
                desc_op = DescOperator.NOT_CONTAINS

            case_sensitive = None
            if q.startswith(f"c{quote}"):
                q = q[1:]
                case_sensitive = True

            if q[0] != quote:
                return Err(
                    "Not a desc token (used to filter against a todo's"
                    " description)."
                )

            end_idx = q[1:].find(quote) + 1
            if end_idx == -1:
                return Err("Bad desc token. No ending quote found.")

            assert not q[end_idx + 1 :] or q[end_idx + 1] == " ", (
                "The character after the last quote should be a space."
                f" query={query}"
            )

            filter_value = q[1:end_idx]
            desc_filter = DescFilter(
                value=filter_value,
                op=desc_op,
                case_sensitive=case_sensitive,
            )
            self.desc_filters.append(desc_filter)
            return Ok(q[end_idx + 2 :])

        return parser

    def priority_parser(self, query: str) -> ErisResult[str]:
        """Parser for todo priority ranges."""
        word, *rest = query.split(" ")
        if word[0] != "(" or word[-1] != ")":
            return Err("Not a priority range.")

        for p in word[1:-1].split(","):
            priority: Priority
            if len(p) == 1:
                priority = cast(Priority, p.upper())
                assert (
                    priority in string.ascii_uppercase
                ), f"Bad priority value: {p}"
                self.priorities.append(priority)
            else:
                assert "-" in p, f"Bad priority range (no dash found): {p}"
                p_range = p.upper()
                start_p, end_p = p_range.split("-")
                n = ord(start_p)
                while n <= ord(end_p):
                    priority = cast(Priority, chr(n))
                    assert (
                        priority in string.ascii_uppercase
                    ), f"Bad priority value: {p}"
                    self.priorities.append(priority)
                    n += 1

        return Ok(" ".join(rest))



"""Contains the Repo class."""

from __future__ import annotations

from dataclasses import dataclass
import operator
from pathlib import Path
from typing import Any, Callable, TypeVar

from eris import ErisResult, Err, Ok
from logrus import Logger
import magodo
from potoroo import Repo, TaggedRepo
from sqlalchemy import func
from sqlmodel import Integer, Session, or_, select
from sqlmodel.sql.expression import SelectOfScalar
from typist import PathLike

from . import db, models
from ._common import NULL_ID
from ._tag import (
    DescOperator,
    GreatTag,
    MetatagOperator,
    MetatagValueType,
    Tag,
)
from ._todo import GreatTodo
from .types import CreateEngineType


logger = Logger(__name__)

SelectOfTodo = SelectOfScalar[models.Todo]
SQLStatementParser = Callable[["SQLTag", SelectOfTodo], SelectOfTodo]
T = TypeVar("T")

# will be populated by the @sql_stmt_parser decorator
SQL_STMT_PARSERS: list[SQLStatementParser] = []


def sql_stmt_parser(parser: SQLStatementParser) -> SQLStatementParser:
    """Decorator that registers statement parsers for SQLTag class."""
    SQL_STMT_PARSERS.append(parser)
    return parser


class SQLRepo(TaggedRepo[str, GreatTodo, GreatTag]):
    """Repo that stores Todos in sqlite database."""

    def __init__(
        self,
        url: str,
        *,
        engine_factory: CreateEngineType = db.create_cached_engine,
        verbose: int = 0,
    ) -> None:
        self.url = url

        # create the Engine object
        kwargs = {}
        if verbose > 2:
            kwargs["echo"] = True
        self.engine = engine_factory(url, **kwargs)

    def add(self, todo: GreatTodo, /, *, key: str = None) -> ErisResult[str]:
        """Adds a new Todo to the DB.

        Returns a unique identifier that has been associated with this Todo.
        """
        with Session(self.engine) as session:
            mtodo = todo.to_model(session, key=key)
            session.add(mtodo)
            session.commit()
            session.refresh(mtodo)

        return Ok(str(mtodo.id))

    def get(self, key: str) -> ErisResult[GreatTodo | None]:
        """Retrieve a Todo from the DB."""
        with Session(self.engine) as session:
            stmt = select(models.Todo).where(models.Todo.id == int(key))
            results = session.exec(stmt)
            mtodo = results.first()
            if mtodo is None:
                return Ok(None)
            else:
                todo = GreatTodo.from_model(mtodo)
                return Ok(todo)

    def remove(self, key: str) -> ErisResult[GreatTodo | None]:
        """Remove a Todo from the DB."""
        with Session(self.engine) as session:
            stmt = select(models.Todo).where(models.Todo.id == int(key))
            results = session.exec(stmt)
            mtodo = results.first()
            if mtodo is None:
                return Ok(None)
            else:
                todo = GreatTodo.from_model(mtodo)

                for mlink in mtodo.metatag_links:
                    delete_metatag = len(mlink.metatag.links) == 1
                    session.delete(mlink)
                    if delete_metatag:
                        session.delete(mlink.metatag)
                    session.commit()

                for mtodo_tags in [
                    mtodo.contexts,
                    mtodo.epics,
                    mtodo.projects,
                ]:
                    for tag in mtodo_tags:  # type: ignore[attr-defined]
                        if len(tag.todos) == 1:
                            session.delete(tag)
                            session.commit()

                session.delete(mtodo)
                session.commit()
                return Ok(todo)

    def get_by_tag(self, tag: GreatTag) -> ErisResult[list[GreatTodo]]:
        """Get Todo(s) from DB by using a tag."""
        todos: list[GreatTodo] = []
        found_mtodo_ids: set[int] = set()
        with Session(self.engine) as session:
            for child_tag in tag.tags:
                stmt = SQLTag(session, child_tag).to_stmt()

                for mtodo in session.exec(stmt).all():
                    if mtodo.id not in found_mtodo_ids:
                        assert mtodo.id is not None, (
                            "All of these Todo models are being pulled from a"
                            " SELECT statement, so they should already have an"
                            " 'id' field."
                        )
                        found_mtodo_ids.add(mtodo.id)
                        todo = GreatTodo.from_model(mtodo)
                        todos.append(todo)
        return Ok(todos)

    def remove_by_tag(self, tag: GreatTag) -> ErisResult[list[GreatTodo]]:
        """Removes Todo(s) from DB by using a tag."""
        removed_todos = self.get_by_tag(tag).unwrap()
        for todo in removed_todos:
            self.remove(todo.ident).unwrap()
        return Ok(removed_todos)

    def all(self) -> ErisResult[list[GreatTodo]]:
        """Returns all Todos contained in the underlying SQL database."""
        todos = []
        with Session(self.engine) as session:
            stmt = select(models.Todo)
            results = session.exec(stmt)
            for mtodo in results.all():
                todo = GreatTodo.from_model(mtodo)
                todos.append(todo)
        return Ok(todos)


@dataclass(frozen=True)
class SQLTag:
    """Wrapper around Tag objects that helps build SQL statements."""

    session: Session
    tag: Tag

    def to_stmt(self) -> SelectOfTodo:
        """Constructs a SQL statement from the provided Tag object."""
        stmt = select(models.Todo)
        for parse_stmt in SQL_STMT_PARSERS:
            stmt = parse_stmt(self, stmt)
        return stmt

    @sql_stmt_parser
    def done_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for done status (i.e. 'x' or 'o')."""
        if self.tag.done is not None:
            stmt = stmt.where(models.Todo.done == self.tag.done)
        return stmt

    @sql_stmt_parser
    def prefix_tag_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for prefix tags (e.g. '@home' or '+greatday')."""
        for prefix_tag_list, link_model, model in [
            (self.tag.contexts, models.ContextLink, models.Context),
            (self.tag.epics, models.EpicLink, models.Epic),
            (self.tag.projects, models.ProjectLink, models.Project),
        ]:
            for prefix_tag in prefix_tag_list:
                if prefix_tag.startswith("-"):
                    name = prefix_tag[1:]
                    op = models.Todo.id.not_in  # type: ignore[union-attr]
                else:
                    name = prefix_tag
                    op = models.Todo.id.in_  # type: ignore[union-attr]

                subquery = (
                    select(models.Todo.id)
                    .join(link_model)
                    .join(model)
                    .where(model.name == name)
                )
                stmt = stmt.where(op(subquery))
        return stmt

    @sql_stmt_parser
    def priority_range_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for priority range (e.g. '(a-c)')."""
        if self.tag.priorities:
            stmt = stmt.where(
                or_(models.Todo.priority == p for p in self.tag.priorities)
            )
        return stmt

    @sql_stmt_parser
    def desc_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for todo description (e.g. '"foo"' or '!"bar"')"""
        for desc_filter in self.tag.desc_filters:
            case_sensitive = desc_filter.case_sensitive
            if case_sensitive is None:
                case_sensitive = not bool(desc_filter.value.islower())

            like_arg = f"%{desc_filter.value}%"
            op_arg: Any
            if case_sensitive:
                cond = models.Todo.desc.like(like_arg)  # type: ignore[attr-defined]
                subquery = select(models.Todo.id, models.Todo.desc).where(cond)
                id_list: list[int] = []
                for ID, desc in self.session.exec(subquery).all():
                    assert (
                        ID is not None
                    ), "The DB shouldn't contain todos without an ID, right?"
                    if desc_filter.value in desc:
                        id_list.append(ID)
                op_map: dict[DescOperator, Any] = {
                    DescOperator.CONTAINS: models.Todo.id.in_,  # type: ignore[union-attr]
                    DescOperator.NOT_CONTAINS: models.Todo.id.not_in,  # type: ignore[union-attr]
                }
                op = op_map[desc_filter.op]
                op_arg = id_list
            else:
                op_map = {
                    DescOperator.CONTAINS: models.Todo.desc.ilike,  # type: ignore[attr-defined]
                    DescOperator.NOT_CONTAINS: models.Todo.desc.not_ilike,  # type: ignore[attr-defined]
                }
                op = op_map[desc_filter.op]
                op_arg = like_arg

            stmt = stmt.where(op(op_arg))
        return stmt

    @sql_stmt_parser
    def date_range_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for create/done dates (e.g. '^2000-01-01' or '$5d:0d')."""
        for date_range_list, model_date in [
            (self.tag.create_date_ranges, models.Todo.create_date),
            (self.tag.done_date_ranges, models.Todo.done_date),
        ]:
            for date_range in date_range_list:
                end_date = date_range.end or date_range.start
                stmt = (
                    stmt.where(model_date is not None)
                    .where(model_date >= date_range.start)  # type: ignore[operator]
                    .where(model_date <= end_date)  # type: ignore[operator]
                )
        return stmt

    @sql_stmt_parser
    def metatag_parser(self, stmt: SelectOfTodo) -> SelectOfTodo:
        """Parser for metatags (e.g. 'due<=0d')."""
        comp_op_map = {
            MetatagOperator.EQ: operator.eq,
            MetatagOperator.NE: operator.ne,
            MetatagOperator.LT: operator.lt,
            MetatagOperator.GT: operator.gt,
            MetatagOperator.LE: operator.le,
            MetatagOperator.GE: operator.ge,
        }
        for mfilter in self.tag.metatag_filters:
            if mfilter.key == "id" and mfilter.op not in [
                MetatagOperator.EXISTS,
                MetatagOperator.NOT_EXISTS,
            ]:
                comp_op = comp_op_map[mfilter.op]
                stmt = stmt.where(
                    comp_op(_col_to_int(models.Todo.id), int(mfilter.value))
                )
                continue

            subquery = (
                select(models.Todo.id)
                .join(models.MetatagLink)
                .join(models.Metatag)
                .where(models.Metatag.name == mfilter.key)
            )

            op = models.Todo.id.in_  # type: ignore[union-attr]
            if mfilter.op == MetatagOperator.EXISTS:
                pass
            elif mfilter.op == MetatagOperator.NOT_EXISTS:
                op = models.Todo.id.not_in  # type: ignore[union-attr]
            else:
                comp_op = comp_op_map[mfilter.op]

                value_type_map: dict[
                    MetatagValueType,
                    tuple[Callable[[Any], Any], Callable[[Any], Any]],
                ] = {
                    MetatagValueType.DATE: (func.date, magodo.dates.to_date),
                    MetatagValueType.INTEGER: (_col_to_int, int),
                    MetatagValueType.STRING: (_noop, _noop),
                }
                cast_model, cast_value = value_type_map[mfilter.value_type]
                subquery = subquery.where(
                    comp_op(
                        cast_model(models.MetatagLink.value),
                        cast_value(mfilter.value),
                    )
                )

            stmt = stmt.where(op(subquery))

        return stmt


def _noop(value: T) -> T:
    """A function that does nothing."""
    return value


def _col_to_int(value: Any) -> Any:
    """Casts SQL table's column to integer."""
    return func.cast(value, Integer)


class FileRepo(Repo[str, GreatTodo]):
    """Repo that stores Todos on disk."""

    def __init__(self, path: PathLike) -> None:
        self.path = Path(path)

    def add(self, todo: GreatTodo, /, *, key: str = None) -> ErisResult[str]:
        """Write a new Todo to disk.

        Returns a unique identifier that has been associated with this Todo.
        """
        if key is None:
            assert todo.ident != NULL_ID, (
                "Every todo added to a file should already have an ID assigned"
                " to it."
            )
            key = todo.ident

        all_todos: list[GreatTodo] = [todo]

        if self.path.exists():
            todos = _todos_from_path(self.path)
            all_todos.extend(todos)
        else:
            self.path.parent.mkdir(parents=True, exist_ok=True)

        with self.path.open("w") as f:
            f.write("\n".join(t.to_line() for t in sorted(all_todos)))

        return Ok(key)

    def get(self, key: str) -> ErisResult[GreatTodo | None]:
        """Retrieve a Todo from disk."""
        for line in self.path.read_text().split("\n"):
            if f"id:{key}" in line.strip().split(" "):
                todo = GreatTodo.from_line(line).unwrap()
                return Ok(todo)
        return Ok(None)

    def remove(self, key: str) -> ErisResult[GreatTodo | None]:
        """Remove a Todo from disk."""
        new_lines: list[str] = []

        todo: GreatTodo | None = None
        for line in self.path.read_text().split("\n"):
            for word in line.strip().split(" "):
                if word == f"id:{key}":
                    todo = GreatTodo.from_line(line).unwrap()
                    break
            else:
                new_lines.append(line)

        self.path.write_text("\n".join(new_lines))

        return Ok(todo)

    def all(self) -> ErisResult[list[GreatTodo]]:
        """Retreive all Todos stored on disk."""
        todos = _todos_from_path(self.path)
        return Ok(todos)


def _todos_from_path(path: PathLike) -> list[GreatTodo]:
    path = Path(path)
    todos: list[GreatTodo] = []
    for line in path.read_text().split("\n"):
        todo_result = GreatTodo.from_line(line)
        if not isinstance(todo_result, Err):
            todos.append(todo_result.ok())
    return todos



"""Contains custom magodo Todo spells used by greatday."""

from __future__ import annotations

import datetime as dt
from typing import Final, Iterable, List

from logrus import Logger
import magodo
from magodo.types import LineSpell, T, TodoSpell
from metaman import register_function_factory

from ._common import drop_word_if_startswith, drop_words, todo_prefixes
from ._dates import (
    RELATIVE_DATE_METATAGS,
    dt_from_date_and_hhmm,
    get_relative_date,
    matches_date_fmt,
    matches_relative_date_fmt,
)


logger = Logger(__name__)

# priority that indicates that a todo is "in progress"
IN_PROGRESS_PRIORITY: Final = "D"

# initialize decorators to register spell functions
GREAT_PRE_TODO_SPELLS: List[TodoSpell] = []
pre_todo_spell = register_function_factory(GREAT_PRE_TODO_SPELLS)

GREAT_TODO_SPELLS: List[TodoSpell] = []
todo_spell = register_function_factory(GREAT_TODO_SPELLS)

GREAT_POST_TODO_SPELLS: List[TodoSpell] = []
post_todo_spell = register_function_factory(GREAT_POST_TODO_SPELLS)

GREAT_TO_LINE_SPELLS: List[LineSpell] = []
to_line_spell = register_function_factory(GREAT_TO_LINE_SPELLS)

GREAT_FROM_LINE_SPELLS: List[LineSpell] = []
from_line_spell = register_function_factory(GREAT_FROM_LINE_SPELLS)


###############################################################################
# pre-todo spells | First, all PRE todo spells are cast...
###############################################################################
@pre_todo_spell
def x_points(todo: T) -> T:
    """Handles metatags of the form 'x:N' at the start of a todo line."""
    x = todo.metadata.get("x")
    if not x or len(x) >= 4:
        return todo

    if not todo.desc.startswith("x:"):
        return todo

    metadata = dict(todo.metadata.items())
    points = metadata["x"]
    del metadata["x"]
    metadata["p"] = points

    desc = todo.desc
    desc = drop_word_if_startswith(desc, "p:")
    desc_words = desc.split(" ")
    desc_words.pop(0)  # x:HHMM

    if matches_date_fmt(desc_words[0]):
        create_date = magodo.dates.to_date(desc_words.pop(0))
    else:
        create_date = None

    if matches_date_fmt(desc_words[0]):
        done_date = create_date
        create_date = magodo.dates.to_date(desc_words.pop(0))
    else:
        done_date = None

    desc = " ".join(desc_words) + f" p:{points}"

    return todo.new(
        create_date=create_date,
        desc=desc,
        done=True,
        done_date=done_date,
        metadata=metadata,
    )


###############################################################################
# normal todo spells | Then all NORMAL todo spells are cast...
###############################################################################
@todo_spell
def snooze_spell(todo: T) -> T:
    """Handles the 'snooze' metadata tag."""
    metadata = dict(todo.metadata.items())
    s = metadata.get("s")
    if s is not None:
        metadata["snooze"] = s
        del metadata["s"]

    snooze = metadata.get("snooze")
    if snooze is None:
        return todo

    del metadata["snooze"]
    metadata["due"] = snooze

    return todo.new(metadata=metadata, priority=magodo.DEFAULT_PRIORITY)


@todo_spell
def render_relative_dates(todo: T) -> T:
    """Renders metatags that support relative dates.

    (e.g. 'due:1d' -> 'due:2022-02-16')
    """
    found_tag = False
    desc = todo.desc
    metadata = dict(todo.metadata.items())

    for key in RELATIVE_DATE_METATAGS:
        value = todo.metadata.get(key)
        if not value:
            continue

        if not matches_relative_date_fmt(value):
            continue

        found_tag = True

        value_date = get_relative_date(value)
        new_value = magodo.dates.from_date(value_date)
        metadata[key] = new_value

        desc = drop_word_if_startswith(desc, key + ":")

    if not found_tag:
        return todo

    return todo.new(desc=desc, metadata=metadata)


@todo_spell
def due_context_spell(todo: T) -> T:
    """Converts @due context into 'due' metatag."""
    if "due" not in todo.contexts:
        return todo

    today = dt.date.today()

    contexts = [ctx for ctx in todo.contexts if ctx != "due"]
    desc = drop_words(todo.desc, "@due")
    metadata = dict(todo.metadata.items())
    metadata["due"] = magodo.dates.from_date(today)
    return todo.new(desc=desc, contexts=contexts, metadata=metadata)


@todo_spell
def appt_todos(todo: T) -> T:
    """Adds priority of (C) or (T) to todos with an appt:HHMM tag."""
    appt = todo.metadata.get("appt")
    if not appt:
        return todo

    if todo.done or todo.done_date:
        return todo

    due = todo.metadata.get("due")
    if due is None:
        return todo

    if not matches_date_fmt(due):
        return todo

    today = dt.date.today()
    if magodo.dates.to_date(due) > today:
        return todo

    now = dt.datetime.now()
    appt_dt = dt_from_date_and_hhmm(today, appt)
    if appt_dt < now + dt.timedelta(minutes=30):
        priority = "C"
    else:
        priority = "T"

    return todo.new(priority=priority)


@todo_spell
def in_progress_priority_spell(todo: T) -> T:
    """Handles todos with the in-prigress priority."""
    start = todo.metadata.get("start")
    if todo.priority != IN_PROGRESS_PRIORITY:
        if start:
            metadata = dict(todo.metadata.items())
            del metadata["start"]
            desc = drop_word_if_startswith(todo.desc, "start:")
            return todo.new(desc=desc, metadata=metadata)
        else:
            return todo

    if start:
        return todo

    now = dt.datetime.now()
    start = f"{now.hour:0>2}{now.minute:0>2}"

    metadata = dict(todo.metadata.items())
    metadata["start"] = start
    desc = todo.desc + f" start:{start}"

    return todo.new(desc=desc, metadata=metadata)


###############################################################################
# post-todo spells | Lastly, all POST todo spells are cast...
###############################################################################
@post_todo_spell
def remove_priorities(todo: T) -> T:
    """Remove priorities for done todos."""
    if todo.priority == magodo.DEFAULT_PRIORITY:
        return todo

    if not todo.done:
        return todo

    priority = magodo.DEFAULT_PRIORITY
    desc = drop_words(todo.desc, f"({todo.priority})")
    return todo.new(desc=desc, priority=priority)


@post_todo_spell
def group_tags(todo: T) -> T:
    """Spell that organizes Todo tags by grouping them at the end.

    Groups all #epics, @ctxs, +projs, and meta:data at the end of the line.
    """
    if not (todo.epics or todo.contexts or todo.projects or todo.metadata):
        return todo

    def all_words_are_tags(words: Iterable[str]) -> bool:
        """Returns True if all `words` are special words."""
        return all(magodo.tags.is_any_tag(w) for w in words)

    all_words = [w for w in todo.desc.split(" ") if w != "|"]
    regular_words: list[str] = []
    for i, word in enumerate(all_words[:]):
        if not word:
            continue

        all_prev_words_are_tags = all_words_are_tags(all_words[:i])
        all_next_words_are_tags = all_words_are_tags(all_words[i + 1 :])
        is_edge_tag = all_next_words_are_tags or all_prev_words_are_tags

        if (
            magodo.tags.is_metadata_tag(word)
            and word[-1] in magodo.PUNCTUATION
        ):
            return todo

        if magodo.tags.is_any_prefix_tag(word) and (
            word[-1] in magodo.PUNCTUATION or not all_next_words_are_tags
        ):
            if regular_words:
                regular_words.append(word[1:])
            continue

        if magodo.tags.is_any_prefix_tag(word) and is_edge_tag:
            continue

        if magodo.tags.is_metadata_tag(word):
            continue

        regular_words.append(word)

    desc = " ".join(regular_words).strip()
    space = ""
    if regular_words:
        desc += " |"
        space = " "

    if todo.epics:
        desc += space + " ".join(
            magodo.tags.EPIC_PREFIX + epic for epic in sorted(todo.epics)
        )
        space = " "

    if todo.contexts:
        desc += space + " ".join(
            magodo.tags.CONTEXT_PREFIX + ctx for ctx in sorted(todo.contexts)
        )
        space = " "

    if todo.projects:
        desc += space + " ".join(
            magodo.tags.PROJECT_PREFIX + ctx for ctx in sorted(todo.projects)
        )
        space = " "

    if todo.metadata:
        desc += space + " ".join(
            f"{k}:{v}" for (k, v) in sorted(todo.metadata.items())
        )
        space = " "

    return todo.new(desc=desc)


###############################################################################
# to-line spells | Called on lines produced by `Todo.to_line()`
###############################################################################
@to_line_spell
def add_o_prefix(line: str) -> str:
    """Adds the 'o ' prefix to the Todo line."""
    if line.startswith(todo_prefixes()):
        return line

    return "o " + line


@to_line_spell
def add_x_prefix(line: str) -> str:
    """Adds the 'x:HHMM ' prefix to the Todo line (when done)."""
    if not line.startswith("x "):
        return line

    words = line.split(" ")[1:]
    for i, word in enumerate(words[:]):
        if word.startswith("dtime:"):
            del words[i]
            dtime = word.split(":")[1]
            break
    else:
        return line

    rest = " ".join(words)
    return f"x:{dtime} {rest}"


###############################################################################
# from-line spells | Called on lines consumed by `Todo.from_line()`
###############################################################################
@from_line_spell
def remove_o_prefix(line: str) -> str:
    """Removes the 'o ' prefix from the Todo line."""
    if not line.startswith("o "):
        return line

    return line[len("o ") :]


@from_line_spell
def remove_x_prefix(line: str) -> str:
    """Removes the 'x:HHMM ' prefix from the Todo line."""
    if not line.startswith("x:"):
        return line

    xhhmm, *words = line.split(" ")
    dtime = xhhmm.split(":")[1]
    if len(dtime) != 4:
        return line

    rest = " ".join(words)
    return f"x {rest} dtime:{dtime}"



"""Contains this project's clack runners."""

from __future__ import annotations

from typing import Final, List

from clack.types import ClackRunner
from logrus import Logger
import metaman

from ._common import CTX_INBOX, drop_words
from ._config import AddConfig, ListConfig, TUIConfig
from ._repo import SQLRepo
from ._tag import GreatTag
from ._todo import GreatTodo
from ._tui import start_textual_app


ALL_RUNNERS: List[ClackRunner] = []
runner = metaman.register_function_factory(ALL_RUNNERS)

logger = Logger(__name__)

CTX_X: Final = "x"


@runner
def run_add(cfg: AddConfig) -> int:
    """Runner for the 'add' subcommand."""
    log = logger.bind_fargs(locals())

    repo = SQLRepo(cfg.database_url, verbose=cfg.verbose)
    todo = GreatTodo.from_line(cfg.todo_line).unwrap()

    x_found = False
    if CTX_X in todo.contexts:
        x_found = True
        desc = drop_words(todo.desc, f"@{CTX_X}")
        contexts = [ctx for ctx in todo.contexts if ctx != CTX_X]
        todo = todo.new(desc=desc, contexts=contexts)

    if cfg.add_inbox_context == "y" or (
        cfg.add_inbox_context == "default"
        and not x_found
        and CTX_INBOX not in todo.contexts
    ):
        contexts = list(todo.contexts) + [CTX_INBOX]
        todo = todo.new(contexts=contexts)

    key = repo.add(todo).unwrap()
    log.info("Added new todo to inbox.", id=repr(key))
    print(todo.to_line())

    return 0


@runner
def run_list(cfg: ListConfig) -> int:
    """Runner for the 'list' subcommand."""
    repo = SQLRepo(cfg.database_url, verbose=cfg.verbose)

    query: str
    if cfg.query is None:
        query = ""
    else:
        query = cfg.query

    tag = GreatTag.from_query(query)
    todos = repo.get_by_tag(tag).unwrap()

    for todo in sorted(todos):
        print(todo.to_line())

    return 0


@runner
def run_tui(cfg: TUIConfig) -> int:
    """Runer for the 'tui' subcommand."""
    start_textual_app(cfg.database_url, verbose=cfg.verbose)
    return 0



"""This script migrates all greatday todos from text files to SQLite DB."""

from __future__ import annotations

import sys

from greatday._common import drop_word_if_startswith
from greatday._repo import FileRepo, SQLRepo


def main() -> int:
    """Main entry point for this script."""
    if len(sys.argv) != 3:
        print(
            "usage: file_repo_to_sql_repo DATA_DIR SQLITEDB", file=sys.stderr
        )
        return 2

    args = list(sys.argv[1:])
    data_dir = args.pop(0)
    sqlite_db = args.pop(0)

    file_repo = FileRepo(data_dir)
    sql_repo = SQLRepo(f"sqlite:///{sqlite_db}")

    for todo in file_repo.todo_group:
        metadata = dict(todo.metadata.items())
        if "id" in metadata:
            metadata["oid"] = metadata["id"]
            del metadata["id"]

        desc = drop_word_if_startswith(todo.desc, "id:")
        new_todo = todo.new(desc=desc, metadata=metadata)
        print(new_todo.to_line())
        sql_repo.add(new_todo).unwrap()
    return 0


if __name__ == "__main__":
    main()



"""Contains greatday's custom Todo types."""

from __future__ import annotations

from typing import Any, cast

from eris import ErisResult, Err, Ok
import magodo
from magodo import MagicTodoMixin
from magodo.types import Priority
from sqlmodel import Session, select

from . import _spells as spells, models
from ._common import NULL_ID, drop_word_if_startswith


_LINE_TO_TODO_CACHE: dict[str, "GreatTodo"] = {}


class GreatTodo(MagicTodoMixin):
    """Custom MagicTodo type used when working with Todos in the daily file."""

    pre_todo_spells = spells.GREAT_PRE_TODO_SPELLS
    todo_spells = spells.GREAT_TODO_SPELLS
    post_todo_spells = spells.GREAT_POST_TODO_SPELLS

    to_line_spells = spells.GREAT_TO_LINE_SPELLS
    from_line_spells = spells.GREAT_FROM_LINE_SPELLS

    @property
    def ident(self) -> str:
        """Returns this Todo's unique identifier."""
        result = self.metadata.get("id", NULL_ID)
        return result

    @classmethod
    def from_line(cls, line: str) -> ErisResult[GreatTodo]:
        """Override's default implementation in order to add caching."""
        todo = _LINE_TO_TODO_CACHE.get(line)
        if todo is None:
            result = super().from_line(line)
            if isinstance(result, Err):
                return result

            todo = result.ok()
            _LINE_TO_TODO_CACHE[line] = todo

        return Ok(todo)

    @classmethod
    def from_model(cls, mtodo: models.Todo) -> GreatTodo:
        """Construct a GreatTodo from a Todo model class."""
        contexts = tuple(ctx.name for ctx in mtodo.contexts)
        epics = tuple(epic.name for epic in mtodo.epics)
        projects = tuple(project.name for project in mtodo.projects)

        metadata = {"id": str(mtodo.id)}
        for mlink in mtodo.metatag_links:
            key = mlink.metatag.name
            value = mlink.value
            metadata[key] = value

        priority = cast(Priority, mtodo.priority)
        todo = magodo.Todo(
            contexts=contexts,
            create_date=mtodo.create_date,
            desc=mtodo.desc,
            done=mtodo.done,
            done_date=mtodo.done_date,
            epics=epics,
            priority=priority,
            projects=projects,
            metadata=metadata,
        )
        return cls(todo)

    def to_model(self, session: Session, key: str = None) -> models.Todo:
        """Converts a GreatTodo into something that the DB can work with."""
        metadata = dict(self.metadata.items())
        id_metatag = metadata.get("id")
        if id_metatag is not None:
            # we don't want to duplicate this in our DB (the primary key will
            # have the same value)
            del metadata["id"]

        if key is None and id_metatag is not None:
            key = id_metatag

        desc = drop_word_if_startswith(self.desc, "id:")
        mtodo_kwargs: dict[str, Any] = dict(
            create_date=self.create_date,
            desc=desc,
            done=self.done,
            done_date=self.done_date,
            priority=self.priority,
        )

        stmt: Any
        if key is None:
            mtodo = models.Todo(**mtodo_kwargs)
        else:
            stmt = select(models.Todo).where(models.Todo.id == int(key))
            results = session.exec(stmt)
            maybe_mtodo = results.first()
            if maybe_mtodo is None:
                mtodo = models.Todo(**mtodo_kwargs)
                mtodo.id = int(key)
            else:
                mtodo = maybe_mtodo
                for k, v in mtodo_kwargs.items():
                    setattr(mtodo, k, v)

        for attr, tag_model in [
            ("contexts", models.Context),
            ("epics", models.Epic),
            ("projects", models.Project),
        ]:
            self_tag_list = getattr(self, attr)
            model_tag_list = []
            for tag_name in self_tag_list:
                stmt = select(tag_model).where(tag_model.name == tag_name)
                results = session.exec(stmt)
                tag = results.first()
                if tag is None:
                    tag = tag_model(name=tag_name)

                model_tag_list.append(tag)
            setattr(mtodo, attr, model_tag_list)

        metatag_links = []
        for k, v in metadata.items():
            stmt = select(models.Metatag).where(models.Metatag.name == k)
            results = session.exec(stmt)
            metatag = results.first()
            if metatag is None:
                metatag = models.Metatag(name=k)

            stmt = (
                select(models.MetatagLink)
                .where(models.MetatagLink.todo_id == mtodo.id)
                .where(models.MetatagLink.metatag_id == metatag.id)
            )
            results = session.exec(stmt)
            mlink = results.first()

            if mlink is None:
                mlink = models.MetatagLink(
                    metatag=metatag, todo=mtodo, value=v
                )

            metatag_links.append(mlink)

        mtodo.metatag_links = metatag_links
        return mtodo



"""Contains the GreatSession class."""

from __future__ import annotations

import datetime as dt
import os
from pathlib import Path
import string
import tempfile
from types import TracebackType
from typing import Type, cast

from logrus import Logger
import magodo
from magodo.types import Priority
from potoroo import Repo, UnitOfWork

from ._common import NULL_ID
from ._dates import get_relative_date
from ._repo import FileRepo, SQLRepo
from ._tag import GreatTag
from ._todo import GreatTodo


logger = Logger(__name__)


class GreatSession(UnitOfWork[FileRepo]):
    """Each time todos are opened in an editor, a new session is created."""

    def __init__(
        self,
        db_url: str,
        tag: GreatTag = None,
        *,
        name: str = None,
        verbose: int = 0,
    ) -> None:
        self.db_url = db_url

        prefix = None if name is None else f"{name}."
        _, temp_path = tempfile.mkstemp(prefix=prefix, suffix=".txt")
        self.path = Path(temp_path)

        self._master_repo = SQLRepo(self.db_url, verbose=verbose)
        self._key_to_old_todo = {}
        if tag is not None:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            with self.path.open("a+") as f:
                for todo in sorted(self._master_repo.get_by_tag(tag).unwrap()):
                    f.write(todo.to_line() + "\n")
                    self._key_to_old_todo[todo.ident] = todo

        # will be accessed via `self.repo` from this point forward
        self._repo = FileRepo(self.path)

    def __enter__(self) -> GreatSession:
        """Called before entering a GreatSession with-block."""
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Called before exiting a GreatSession with-block."""
        del exc_type
        del exc_value
        del traceback

        os.unlink(self.path)

    def commit(self) -> None:
        """Commit our changes.

        We achieve this by copying the contents of the backup file created on
        instantiation back to the original.
        """
        removed_todo_keys = list(self._key_to_old_todo.keys())
        new_todos = {}
        for todo in self.repo.all().unwrap():
            key = todo.ident
            if key in removed_todo_keys:
                removed_todo_keys.remove(key)

            old_todo = self._key_to_old_todo.get(key)
            if key == NULL_ID:
                logger.info("New todo was added while editing?", todo=todo)
                key = self._master_repo.add(todo).unwrap()
                new_todos[key] = todo
            elif todo != old_todo:
                _commit_todo_changes(self._master_repo, todo, old_todo)

        if new_todos:
            # HACK: Removes all new todos by assuming that new todos will not
            # have been assigned an ID yet.
            old_lines = self.path.read_text().split("\n")
            self.path.write_text(
                "\n".join(line for line in old_lines if " id:" in line)
            )

        for key, todo in new_todos.items():
            self._key_to_old_todo[key] = todo
            self.repo.add(todo, key=key)

        for key in removed_todo_keys:
            removed_todo = self._master_repo.remove(key).unwrap()
            if removed_todo is not None:
                logger.info("Todo has been deleted.", todo=removed_todo)
                del self._key_to_old_todo[removed_todo.ident]

    def rollback(self) -> None:
        """Revert any changes made while in this GreatSession's with-block."""

    @property
    def repo(self) -> FileRepo:
        """Returns the GreatRepo object associated with this GreatSession."""
        return self._repo


def _commit_todo_changes(
    repo: Repo[str, GreatTodo], todo: GreatTodo, old_todo: GreatTodo | None
) -> None:
    """Updates todo in repo.

    This function also handles recurring todos (i.e. todos with the
    'recur' metatag).
    """
    recur = todo.metadata.get("recur")
    until = todo.metadata.get("until")
    expired = bool(
        todo.done_date
        and until
        and magodo.dates.to_date(until) <= todo.done_date
    )
    if (
        old_todo
        and todo.done_date
        and not old_todo.done_date
        and recur
        and not expired
    ):
        next_metadata = dict(todo.metadata.items())

        # set 'prev' and 'xp' metatags for next todo...
        next_metadata["prev"] = next_metadata["id"]
        del next_metadata["id"]
        if next_metadata.get("p"):
            del next_metadata["p"]

        # set 'due' metatag for next todo...
        due = todo.metadata.get("due")
        if recur.islower() or due is None:
            start_date = todo.done_date
        else:
            start_date = magodo.dates.to_date(due)

        next_date = get_relative_date(recur, start_date=start_date)
        next_metadata["due"] = magodo.dates.from_date(next_date)

        # set creation date + clear creation/done time for next todo...
        next_create_date = dt.date.today()
        for key in ["ctime", "dtime"]:
            if key in next_metadata:
                del next_metadata[key]

        # set priority for next todo...
        priority = todo.metadata.get("priority")
        next_priority = magodo.DEFAULT_PRIORITY
        if (
            priority
            and len(priority) == 1
            and priority.upper() in string.ascii_uppercase
        ):
            next_priority = cast(Priority, priority.upper())
        elif priority:
            logger.warning("Bad 'priority' metatag value?", priority=priority)

        # add next todo to repo...
        next_todo = todo.new(
            create_date=next_create_date,
            done=False,
            done_date=None,
            metadata=next_metadata,
            priority=next_priority,
        )
        next_key = repo.add(next_todo).unwrap()

        # add 'next' metatag to old todo...
        metadata = dict(todo.metadata.items())
        metadata["next"] = next_key
        todo = todo.new(metadata=metadata)

    repo.update(todo.ident, todo).unwrap()



"""Tests for greatday's SQLRepo class."""

from __future__ import annotations

from pytest import mark

from greatday._repo import SQLRepo
from greatday._tag import GreatTag
from greatday._todo import GreatTodo

from . import common


params = mark.parametrize


def test_add(sql_repo: SQLRepo) -> None:
    """Tests the SQLRepo.add() method.

    NOTE: Nothing needs to be done here since the repo fixture invokes the
    SQLRepo.add() method for us.
    """
    assert len(common.TODO_LINES) == len(sql_repo.all().unwrap())


@params("key", common.TODO_LINE_IDS)
def test_get_and_remove(sql_repo: SQLRepo, key: str) -> None:
    """Tests the SQLRepo.get() and SQLRepo.remove() methods."""
    todo = sql_repo.get(key).unwrap()
    assert todo == sql_repo.remove(key).unwrap()
    assert len(common.TODO_LINES) == len(sql_repo.all().unwrap()) + 1


@params("key", common.TODO_LINE_IDS)
def test_update(sql_repo: SQLRepo, key: str) -> None:
    """Tests the SQLRepo.update() method."""
    old_todo = GreatTodo.from_line(
        common.TODO_LINES[int(key) - 1] + f" id:{key}"
    ).unwrap()
    todo = GreatTodo.from_line(f"o foobar @foo @bar id:{key}").unwrap()

    actual_old_todo = sql_repo.update(key, todo).unwrap()
    assert old_todo == actual_old_todo

    actual_todo = sql_repo.get(key).unwrap()
    assert todo == actual_todo


@params("query,keys", common.QUERY_TO_TODO_IDS)
def test_get_by_tag(sql_repo: SQLRepo, query: str, keys: list[str]) -> None:
    """Tests the SQLRepo.get_by_tag() method."""
    tag = GreatTag.from_query(query)
    todos = sql_repo.get_by_tag(tag).unwrap()
    assert sorted(todo.ident for todo in todos) == sorted(str(k) for k in keys)


def test_remove_by_tag(sql_repo: SQLRepo) -> None:
    """Tests the SQLRepo.remove_by_tag() method."""
    CTX = "@dev"
    matched_line_count = len(
        [line for line in common.TODO_LINES if CTX in line.split(" ")]
    )
    tag = GreatTag.from_query(CTX)
    removed_todos = sql_repo.remove_by_tag(tag).unwrap()
    assert len(removed_todos) == matched_line_count
    assert (
        len(sql_repo.all().unwrap())
        == len(common.TODO_LINES) - matched_line_count
    )
